use std::mem::transmute;
use std::{
    any::{Any, TypeId},
    collections::{
        HashMap,
        hash_map::{Iter, IterMut},
    },
    fmt::Display,
};

use anyhow::Result;
use destructure_traitobject;
use serde::{Deserialize, Serialize};

use crate::World;

pub trait EntityComponent: Any + Send + Sync {
    fn enable(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn disable(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn name(&self) -> &str;
}

pub trait ComponentDebug {
    fn debug_name() -> &'static str;
}

pub trait ComponentCreateInfo<'de>: Serialize + Deserialize<'de> {
    fn to_string(&self) -> Result<String> {
        Ok(ron::to_string(self)?)
    }

    fn from_str(s: &'de str) -> Option<Self> {
        match ron::from_str(s) {
            Ok(t) => Some(t),
            Err(err) => {
                println!("{:?}", err);

                None
            }
        }
    }
}

pub struct TypeMap {
    map: HashMap<TypeId, Box<dyn EntityComponent>>,
}

impl Default for TypeMap {
    fn default() -> Self {
        Self {
            map: HashMap::new(),
        }
    }
}

impl TypeMap {
    pub fn insert<T: EntityComponent + ComponentDebug>(&mut self, value: T) -> Option<T> {
        self.map
            .insert(TypeId::of::<T>(), Box::new(value))
            .map(|any| *Self::downcast_unchecked(any))
    }

    pub fn insert_type(
        &mut self,
        type_id: TypeId,
        component: Box<dyn EntityComponent>,
    ) -> Option<Box<dyn EntityComponent>> {
        self.map.insert(type_id, component)
    }

    pub fn remove<T: EntityComponent>(&mut self) -> Option<T> {
        self.remove_by_type_id(&TypeId::of::<T>())
            .map(|any| *Self::downcast_unchecked(any))
    }

    pub fn remove_by_type_id(&mut self, type_id: &TypeId) -> Option<Box<dyn EntityComponent>> {
        self.map.remove(type_id)
    }

    pub fn get<T: EntityComponent + ComponentDebug>(
        &self,
    ) -> std::result::Result<&T, ComponentNotFoundError> {
        self.map
            .get(&TypeId::of::<T>())
            .map(|any| Self::downcast_ref_unchecked(any))
            .ok_or_else(ComponentNotFoundError::component::<T>)
    }

    pub fn get_by_type_id(
        &self,
        type_id: &TypeId,
    ) -> std::result::Result<&Box<dyn EntityComponent>, ComponentNotFoundError> {
        self.map
            .get(type_id)
            .ok_or_else(|| ComponentNotFoundError::type_id(*type_id))
    }

    pub fn get_mut<T: EntityComponent + ComponentDebug>(
        &mut self,
    ) -> std::result::Result<&mut T, ComponentNotFoundError> {
        self.map
            .get_mut(&TypeId::of::<T>())
            .map(|any| Self::downcast_mut_unchecked(any))
            .ok_or_else(ComponentNotFoundError::component::<T>)
    }

    pub fn get_mut_by_type_id(
        &mut self,
        type_id: &TypeId,
    ) -> std::result::Result<&mut Box<dyn EntityComponent>, ComponentNotFoundError> {
        self.map
            .get_mut(type_id)
            .ok_or_else(|| ComponentNotFoundError::type_id(*type_id))
    }

    pub fn contains<T: EntityComponent>(&self) -> bool {
        self.contains_type_id(&TypeId::of::<T>())
    }

    pub fn contains_type_id(&self, type_id: &TypeId) -> bool {
        self.map.contains_key(type_id)
    }

    pub fn multi_mut(&mut self) -> MultiMut<'_> {
        MultiMut::new(&mut self.map)
    }

    pub fn iter(&self) -> Iter<'_, TypeId, Box<dyn EntityComponent>> {
        self.map.iter()
    }

    pub fn iter_mut(&mut self) -> IterMut<'_, TypeId, Box<dyn EntityComponent>> {
        self.map.iter_mut()
    }

    fn downcast_unchecked<T: EntityComponent>(boxed: Box<dyn EntityComponent>) -> Box<T> {
        unsafe { Box::from_raw(Box::into_raw(boxed) as *mut T) }
    }

    pub fn downcast_ref_unchecked<T: EntityComponent>(boxed_ref: &Box<dyn EntityComponent>) -> &T {
        unsafe {
            let ptr_to_ptr: *const *const T =
                transmute(destructure_traitobject::data(boxed_ref as *const _));

            &**ptr_to_ptr
        }
    }

    pub fn downcast_mut_unchecked<T: EntityComponent>(
        boxed_ref: &mut Box<dyn EntityComponent>,
    ) -> &mut T {
        unsafe {
            let ptr_to_ptr: *mut *mut T =
                transmute(destructure_traitobject::data(boxed_ref as *mut _));

            &mut **ptr_to_ptr
        }
    }
}

/// Allows mutable access to multiple components at once
pub struct MultiMut<'a> {
    map: &'a mut HashMap<TypeId, Box<dyn EntityComponent>>,
    buffer: Vec<*mut Box<dyn EntityComponent>>,
}

impl<'a> MultiMut<'a> {
    fn new(map: &'a mut HashMap<TypeId, Box<dyn EntityComponent>>) -> Self {
        MultiMut {
            map,
            buffer: Vec::new(),
        }
    }

    /// Returns requested type on success
    pub fn get<T: EntityComponent + ComponentDebug>(
        &mut self,
    ) -> std::result::Result<&'a mut T, ComponentNotFoundError> {
        self.get_by_type_id(&TypeId::of::<T>())
            .map(|component| TypeMap::downcast_mut_unchecked(component))
            .map_err(|_| ComponentNotFoundError::component::<T>())
    }

    /// Returns requested type behind this type id on success
    pub fn get_by_type_id(
        &mut self,
        type_id: &TypeId,
    ) -> std::result::Result<&'a mut Box<dyn EntityComponent>, ComponentNotFoundError> {
        match self.map.get_mut(type_id) {
            Some(v) => {
                let ptr = v as *mut _;

                match self.buffer.iter().find(|v| **v == ptr) {
                    Some(_) => {
                        panic!("This key has already been borrowed!");
                    }
                    None => {
                        self.buffer.push(ptr);
                    }
                }

                let t: Option<&'a mut Box<dyn EntityComponent>> = unsafe { transmute(ptr) };

                t.ok_or_else(|| ComponentNotFoundError::type_id(*type_id))
            }
            None => Err(ComponentNotFoundError::type_id(*type_id)),
        }
    }

    /// # Safety
    ///
    /// use this only when there are no references left
    pub unsafe fn clear_all_usages(&mut self) {
        self.buffer.clear();
    }
}

#[derive(Debug)]
pub enum ComponentRequestType {
    TypeId(TypeId),
    ComponentName(&'static str),
}

#[derive(Debug)]
pub struct ComponentNotFoundError {
    pub request_type: ComponentRequestType,
}

impl ComponentNotFoundError {
    fn type_id(type_id: TypeId) -> Self {
        Self {
            request_type: ComponentRequestType::TypeId(type_id),
        }
    }

    fn component<T: ComponentDebug>() -> Self {
        Self {
            request_type: ComponentRequestType::ComponentName(T::debug_name()),
        }
    }
}

impl Display for ComponentNotFoundError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Entity Component ({:?}) not found!", self.request_type)
    }
}

impl std::error::Error for ComponentNotFoundError {}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct Test {
    x: u32,
    y: u32,
    z: u32,
}

impl EntityComponent for Test {
    fn name(&self) -> &str {
        "Test"
    }
}

impl ComponentDebug for Test {
    fn debug_name() -> &'static str {
        "Test"
    }
}

impl PartialEq for Test {
    fn eq(&self, other: &Self) -> bool {
        self.x == other.x && self.y == other.y && self.z == other.z
    }
}

#[test]
fn verify_multi_mut() {
    let mut map = TypeMap::default();

    map.insert(Test::default());

    // test it multiple times, just for sanity
    for _ in 0..10 {
        let test = map.get::<Test>().unwrap().clone();

        let mut multi_mut = map.multi_mut();

        let multi_mut_test = multi_mut.get::<Test>().unwrap().clone();

        assert_eq!(test, multi_mut_test);
    }
}

#[test]
fn verify_insert() {
    let mut map = TypeMap::default();

    let reference = Test { x: 5, y: 20, z: 30 };

    map.insert(reference.clone());
    let old = map.insert(reference.clone()).unwrap();

    assert_eq!(old, reference);
}

#[test]
fn verify_get() {
    let mut map = TypeMap::default();

    let reference = Test { x: 5, y: 20, z: 30 };

    map.insert(reference.clone());

    assert_eq!(map.get::<Test>().unwrap().clone(), reference);
}

#[test]
fn verify_get_mut() {
    let mut map = TypeMap::default();

    let reference = Test { x: 5, y: 20, z: 30 };

    map.insert(reference.clone());

    assert_eq!(map.get_mut::<Test>().unwrap().clone(), reference);
}
