use std::{fmt, ptr};

#[derive(Clone)]
pub struct UnsafeComponentStore<T> {
    ptr: *mut T,
}

impl<T> UnsafeComponentStore<T> {
    /// # Safety
    ///
    /// Creates structure with null pointer. Should never be called like this.
    pub unsafe fn empty() -> Self {
        Self {
            ptr: ptr::null_mut(),
        }
    }

    /// # Safety
    ///
    /// Returns mutable reference to the inner pointer. Use always with caution
    /// and make 100% sure that the actual struct is still present.
    #[allow(clippy::mut_from_ref)]
    pub unsafe fn as_mut(&self) -> &mut T {
        assert_ne!(
            self.ptr,
            ptr::null_mut(),
            "Called UnsafeComponentStore while being empty"
        );

        unsafe { self.ptr.as_mut().unwrap() }
    }

    pub fn is_init(&self) -> bool {
        !self.ptr.is_null()
    }
}

impl<T> From<&mut T> for UnsafeComponentStore<T> {
    fn from(ptr: &mut T) -> Self {
        Self { ptr: ptr as *mut T }
    }
}

impl<T> From<&T> for UnsafeComponentStore<T> {
    fn from(ptr: &T) -> Self {
        Self {
            ptr: ptr as *const T as *mut T,
        }
    }
}

impl<T> fmt::Debug for UnsafeComponentStore<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("UnsafeComponentStore")
            .field("ptr", &self.ptr)
            .finish()
    }
}

impl<T> Default for UnsafeComponentStore<T> {
    fn default() -> Self {
        unsafe { Self::empty() }
    }
}

unsafe impl<T> Send for UnsafeComponentStore<T> {}
unsafe impl<T> Sync for UnsafeComponentStore<T> {}
